(* -*- sml -*-
 * input.cpp
 *
 * COPYRIGHT (c) 2001 Bell Labs, Lucent Technologies
 *
 * Input routines for SML/NG documents.  Note that the task specification
 * allows us to do what every we want on bad input, so we cut some corners
 * and do not do error checking.
 *
 * TODO:
 *  merge input operations and parsing to get better cache behavior
 *  remove error checks from tag parsing
 *  switch to unsafe subscripting
 *)

structure Input : sig

    val input : unit -> Document.doc

  end = struct

    structure D = Document
    structure SIO = TextIO.StreamIO
    structure SS = Substring
    structure OU = OptUtil

    structure II = FastIInfer

  (* subscript operations *)
#include "sub.ins"

    datatype char_class
      = EOF
      | BR		(* break between bufers *)
      | LT		(* '<' *)
      | SP		(* space *)
      | CHR

    val ccMap = let
	  val tbl = Array.tabulate(256,
		fn i => if ((0x21 <= i) andalso (i <= 0x7E)) then CHR else BR)
	  in
	    List.app (fn (c, cc) => Array.update(tbl, ord c, cc)) [
		(#"<", LT), (#" ", SP), (#"\n", SP), (#"\t", SP), (#"\r", SP)
	      ];
	    fn c => arraySub(tbl, ord c)
	  end

  (* get the next character and its class *)
    fun getcc ([], i) = (EOF, #" ", ([], i))
      | getcc (data as (buf::r), i) =
	  if (i < size buf)
	    then let
	      val c = stringSub(buf, i)
	      in
		(ccMap c, c, (data, i+1))
	      end
	    else (BR, #" ", (r, 0))

  (* get the next character *)
    fun getc ([], _) = raise Fail "unexpected EOF"
      | getc (data as (buf::r), i) =
	  if (i < size buf)
	    then (stringSub(buf, i), (data, i+1))
	    else getc(r, 0)

  (* peek at the character class of the next character; skipping over BRs *)
    fun peekcc ([], _) = EOF
      | peekcc (data as (buf::r), i) =
	  if (i < size buf)
	    then ccMap(stringSub(buf, i))
	    else peekcc(r, 0)

#include "bits.ins"

  (* markup tags *)
    datatype tag = Font of int | Sz of int | Col of int
    val tagPL = Font pl_num
    val tagB = Font b_num
    val tagI = Font i_num
    val tagTT = Font tt_num
    val tagS = Font s_num
    val tagEM = Font em_num
    val tagU = Font u_num
    val tag0 = Sz 0
    val tag1 = Sz 1
    val tag2 = Sz 2
    val tag3 = Sz 3
    val tag4 = Sz 4
    val tag5 = Sz 5
    val tag6 = Sz 6
    val tag7 = Sz 7
    val tag8 = Sz 8
    val tag9 = Sz 9
    val tagr = Col 0	(* red *)
    val tagg = Col 1	(* green *)
    val tagb = Col 2	(* blue *)
    val tagc = Col 3	(* cyan *)
    val tagm = Col 4	(* magenta *)
    val tagy = Col 5	(* yellow *)
    val tagk = Col 6	(* black *)
    val tagw = Col 7	(* white *)

(***** BEGIN SLOW TAGGING CODE *****)
    val plTest = testBit pl_bit

    fun extendFTagging (bit, tagging as {color, sz, font}) =
	  if (testBit bit font) orelse (plTest font)
	    then tagging
	    else {color = color, sz = sz, font = setBit bit font}

  (* extend a tagging by a tag; the tag is outside the document marked by
   * the tagging.
   *)
    fun extendTagging (Font 0, tagging as {color, sz, font}) = let (* <U> *)
	  val u = projBits u_bits font
	  in
	    if (u = 0w3) orelse plTest font
	      then tagging
	      else {color=color, sz=sz, font=setU(font, u+0w1)}
	  end
      | extendTagging (Font 1, tagging as {color, sz, font}) = (* <EM> *)
	  if testBit font s_bit orelse plTest font
	    then tagging
	    else {color = color, sz = sz, font = flipBit em_bit font}
      | extendTagging (Font 2, tagging as {color, sz, font}) = (* <S> *)
	  if testBit font s_bit orelse plTest font
	    then tagging
	    else {
		color = color, sz = sz,
		font = setBit s_bit (projBits font em_mask)
	      }
      | extendTagging (Font 3, tagging as {color, sz, font}) = (* <TT> *)
	  extendFTagging (tt_bit, tagging)
      | extendTagging (Font 4, tagging) = (* <I> *)
	  extendFTagging (i_bit, tagging)
      | extendTagging (Font 5, tagging) = (* <B> *)
	  extendFTagging (b_bit, tagging)
      | extendTagging (Font 6, tagging as {color, sz, font}) = (* <PL> *)
	(* the PL tag is tricky; it has not effect if pl is already set,
	 * or if all of the  *)
	  if plTest font
	  orelse (projBits (Word.andb(pl_mask, u_mask)) font
	      = Word.andb(pl_mask, u_mask)
	    andalso (projBits u_bits font > 0w0))
	    then tagging
	    else {color = color, sz = sz, font = setBit pl_bit font}
      | extendTagging (Sz n, tagging as {color, sz, font}) =
	  if sz >= 0 then tagging else {color=color, sz=n, font=font}
      | extendTagging (Col c, tagging as {color, sz, font}) =
	  if color >= 0 then tagging else {color=c, sz=sz, font=font}

    fun toTagging b = {color = ~1, sz = ~1, font = b}

  (* create a singleton tagging from a tag *)
    fun tagToTagging (Font 0) = (* <U> *) toTagging (setU(0w0, 0w1))
      | tagToTagging (Font 1) = (* <EM> *) toTagging em_bit
      | tagToTagging (Font 2) = (* <S> *) toTagging s_bit
      | tagToTagging (Font 3) = (* <TT> *) toTagging tt_bit
      | tagToTagging (Font 4) = (* <I> *) toTagging i_bit
      | tagToTagging (Font 5) = (* <B> *) toTagging b_bit
      | tagToTagging (Font 6) = (* <PL> *) toTagging pl_bit
      | tagToTagging (Sz n) = {color = ~1, sz = n, font = 0w0}
      | tagToTagging (Col c) = {color = c, sz = ~1, font = 0w0}

    val slowExtendTagging = extendTagging
    val slowTagToTagging = tagToTagging
(***** END SLOW TAGGING CODE *****)

  (* Contexts describe the current state; for now we are using a record
   * representation, but we can switch to a bit-vector representation with
   * transition tables later.
   *)
    type context = {
	  font : word,	(* U@0..1, EM@2, S@3, TT@4, I@5, B@6 *)
	  sz : int,	(* 0..9; ~1 for unknown *)
	  color : int	(* 0..7; ~1 for unknown *)
	}

    fun nextFontState (font, tag) =
	  stringSubW(arraySub(Tables.contextTbl, tag), font)

(*  *********** BEGIN Checked tagging machinery **************** *)

    fun extendTagging (tag, tagging) = let
#ifndef NDEBUG
	  fun showTagging {color, sz, font} = concat [
		  "c(", Int.toString color, ")s(", Int.toString sz,
		  ")f(", Word.toString font, ")"
		]
	  val tagging1 = slowExtendTagging (tag, tagging)
#endif
	  val tagging2 = OU.collapseTagging(tagToTagging tag, tagging)
	  in
#ifdef NDEBUG
            tagging2
#else
	    if (tagging1 = tagging2)
	      then tagging1
	      else (
		TextIO.output (TextIO.stdErr, concat [
		    "!!! discrepancy in extendTagging!\n",
		    "  t1=", showTagging(tagToTagging tag),
		    ", t2=", showTagging tagging,
		    ", fast=", showTagging tagging2,
		    ", slow=", showTagging tagging1, "\n"
		  ]);
		OS.Process.exit OS.Process.failure)
#endif
	  end

(*  *********** END Checked tagging machinery **************** *)

    datatype next_context = SAME | NEXT of context

  (* this function returns the next context given the current context and a tag.
   * It tests to see if a given tag is a NOP in the current context and returns
   * SAME in that case.
   *)
    fun nextContext (c as {font, sz, color}, tag) = (case tag
	   of (Font i) => let
		val font2 = nextFontState (font, i)
		in
		  if font = font2
		    then SAME
		    else NEXT{font = font2, sz = sz, color = color}
		end
	    | Sz n => if (sz = n)
		then SAME
		else NEXT{font = font, sz=n, color=color}
	    | Col c => if (color = c)
		then SAME
		else NEXT{font = font, sz=sz, color=c}
	  (* end case *))

    fun preserveWS {font, sz, color} = testBit tt_bit font

    val rootContext = {font = 0w0, sz = ~1, color = ~1}

  (* a space context is a projection of the parts of the context that
   * affect whitespace.  We do not explicitly compute the space context,
   * but instead use a full context as its representation.
   * NOTE: the tt field is part of the space context, except that we
   * disable space merging when it is set in the full context, so tt
   * will always be false when this data structure is being used.
   *)
    datatype prev_space_context = NOT_SP | SPC of context

  (* this function is called when a space is parsed to determine if it can be
   * combined with the previous space.
   *)
    local val projU = projBits u_bits in
    fun combineSpace (NOT_SP, _) = false
      | combineSpace (SPC{font, color, sz}, c : context) = let
	  val u = projU font
	  in
	    (u = projU(#font c)) andalso (sz = #sz c)
	    andalso ((u = 0w0) orelse (color = #color c))
	  end
    end

    datatype tag_token = OPEN of tag | CLOSE

 (* get a markup tag having already consumed the "<".  This code accepts
  * tags like "</////TT>".
   *)
    fun getTag strm = let
	  fun get1 (isOpen, strm) = (case getc strm
		 of (#"/", strm2) => get1 (false, strm2)
		  | (#"B", strm2) => getGT (isOpen, tagB, strm2)
		  | (#"E", strm2) => get2 (isOpen, tagEM, #"M", strm2)
		  | (#"I", strm2) => getGT (isOpen, tagI, strm2)
		  | (#"P", strm2) => get2 (isOpen, tagPL, #"L", strm2)
		  | (#"S", strm2) => getGT (isOpen, tagS, strm2)
		  | (#"T", strm2) => get2 (isOpen, tagTT, #"T", strm2)
		  | (#"U", strm2) => getGT (isOpen, tagU, strm2)
		  | (#"0", strm2) => getGT (isOpen, tag0, strm2)
		  | (#"1", strm2) => getGT (isOpen, tag1, strm2)
		  | (#"2", strm2) => getGT (isOpen, tag2, strm2)
		  | (#"3", strm2) => getGT (isOpen, tag3, strm2)
		  | (#"4", strm2) => getGT (isOpen, tag4, strm2)
		  | (#"5", strm2) => getGT (isOpen, tag5, strm2)
		  | (#"6", strm2) => getGT (isOpen, tag6, strm2)
		  | (#"7", strm2) => getGT (isOpen, tag7, strm2)
		  | (#"8", strm2) => getGT (isOpen, tag8, strm2)
		  | (#"9", strm2) => getGT (isOpen, tag9, strm2)
		  | (#"r", strm2) => getGT (isOpen, tagr, strm2)
		  | (#"g", strm2) => getGT (isOpen, tagg, strm2)
		  | (#"b", strm2) => getGT (isOpen, tagb, strm2)
		  | (#"c", strm2) => getGT (isOpen, tagc, strm2)
		  | (#"m", strm2) => getGT (isOpen, tagm, strm2)
		  | (#"y", strm2) => getGT (isOpen, tagy, strm2)
		  | (#"k", strm2) => getGT (isOpen, tagk, strm2)
		  | (#"w", strm2) => getGT (isOpen, tagw, strm2)
		  | _ => raise Fail "bogus markup"
		(* end case *))
	  and get2 (isOpen, tag, c, strm) = let
		val (c2, strm2) = getc strm
		in
		  if (c = c2)
		    then getGT (isOpen, tag, strm2)
		    else raise Fail "bogus markup"
		end
	  and getGT (isOpen, tag, strm) = (case getc strm
		 of (#">", strm2) =>
		      (if isOpen then OPEN tag else CLOSE, strm2)
		  | _ => raise Fail "bogus markup"
		(* end case *))
	  in
	    get1 (true, strm)
	  end
 
    fun getFrag (start, (s::_, i)) = SS.substring(s, start, i - start)
      | getFrag _ = SS.all ""

    local
	fun space_ign (context : context) = let
	    val u = projBits u_bits (#font context)
	    infix ||
	    val op || = Word.orb
	in
	    { color = u = 0w0, sz = false,
	      font = (b_bit || i_bit || s_bit || em_bit) || 0w3
	    }
	end
    in
    fun mkSp context = D.SP{ign = space_ign context}

    fun mkText (body, context) = let
	fun isWhite c = ccMap c = SP
	fun allWhite s = let
	    fun loop i =
		i < 0 orelse
		(isWhite (stringSub (s, i)) andalso loop (i - 1))
	in
	    loop (size s - 1)
	end
    in
	D.TEXT { body = body,
		 ign = if allWhite body then space_ign context
		       else II.i_none }
    end

    end (* local *)

    fun getText (start, (s::_, i), context) =
	  mkText (substring(s, start, i - start), context)

    val space = SS.all " "

    fun mkList [doc] = doc
      | mkList docs = D.LIST{
	    body = docs,
	    ign = II.i_list (map D.ignOf docs)
	  }

    fun mkElem (tags, content) = D.ELEM{
	    tags = tags, content = content,
	    ign = II.i_elem (tags, D.ignOf content)
	  }

    fun mkDoc (_, [], prev) = prev
      | mkDoc (tag, [D.ELEM{tags, content, ...}], prev) =
	    mkElem (extendTagging(tag, tags), content) :: prev
      | mkDoc (tag, [doc], prev) =
	  mkElem (tagToTagging tag, doc) :: prev
      | mkDoc (tag, docs, prev) =
	  mkElem (tagToTagging tag, mkList docs) :: prev

    fun parse (data : string list) = let
	  fun scanContent (context, prevSp, strm) = let
		val preserveWS = preserveWS context
		fun scan (prevSp, strm, body : D.doc list) = let
		      fun scanText (start, strm) = let
			    fun scanTT (start, strm) = (case getcc strm
				   of (CHR, _, strm2) => scanTT (start, strm2)
				    | (SP, _, strm2) => scanTT (start, strm2)
				    | _ => (getText(start, strm, context),
					    strm)
				  (* end case *))
			    fun scanTxt (start, strm, frags) = (case getcc strm
				   of (CHR, _, strm2) =>
					scanTxt (start, strm2, frags)
				    | (SP, _, strm2) => let
					val frag = getFrag (start, strm)
					val strm3 = scanWS strm2
					in
					  if (peekcc strm3 = LT)
					    then let
					      val text =
						    SS.concat(List.rev(frag::frags))
					      in
						(mkText (text, context), strm)
					      end
					    else scanTxt(#2 strm3, strm3,
					      space::frag::frags)
					end
				    | _ => let
					val frag = getFrag (start, strm)
					val text = SS.concat(List.rev(frag::frags))
					in
					  (mkText (text, context), strm)
					end
				  (* end case *))
			    in
			      if preserveWS
				then scanTT (start, strm)
				else scanTxt (start, strm, [])
			    end
		      and scanWS strm = (case getcc strm
			     of (SP, _, strm2) => scanWS strm2
			      | (BR, _, strm2) => scanWS strm2
			      | _ => strm
			    (* end case *))
		      fun scanElement (tag, strm) = (case nextContext(context, tag)
			     of SAME => let
				  val (prevSp, content, strm2) =
					scanContent (context, prevSp, strm)
				  in
				    scan (prevSp, strm2, List.revAppend(content, body))
				  end
			      | NEXT ctx => (case scanContent (ctx, prevSp, strm)
				   of (prevSp, [], strm2) =>
				      (* delete empty tags *)
					scan (prevSp, strm2, body)
				    | (prevSp, content, strm2) =>
					scan (prevSp, strm2,
					  mkDoc(tag, content, body))
				  (* end case *))
			    (* end case *))
		      in
			case getcc strm
			 of (EOF, _, _) => (prevSp, List.rev body, strm)
			  | (BR, _, strm2) => scan (prevSp, strm2, body)
			  | (LT, _, strm2) => (case getTag strm2
			       of (OPEN tag, strm3) => scanElement (tag, strm3)
				| (CLOSE, strm3) => (prevSp, List.rev body, strm3)
			      (* end case *))
			  | (SP, _, strm2) => if preserveWS
			      then let
				val (tok, strm3) = scanText(#2 strm, strm2)
				in
				  scan (NOT_SP, strm3, tok::body)
				end
			      else let
				val strm3 = scanWS strm2
				in
				  if combineSpace (prevSp, context)
				    then scan (prevSp, strm3, body)
				    else scan (
				      SPC context, strm3,
				      mkSp context::body)
				end
			  | (CHR, _, strm2) => let
			      val (tok, strm3) = scanText(#2 strm, strm2)
			      in
				scan (NOT_SP, strm3, tok::body)
			      end
			(* end case *)
		      end
	      in
		scan (prevSp, strm, [])
	      end
	  in
	    scanContent(rootContext, NOT_SP, (data, 0))
	  end

    fun input () = let
	  val inStrm = TextIO.getInstream TextIO.stdIn
	  fun getFile strm = (case SIO.input strm
		 of ("", _) => []
		  | (s, strm) => s :: getFile strm
		(* end case *))
	  val data = getFile inStrm
	  val (_, doc, _) = parse data
	  in
	    mkList doc
	  end

  end
