(* -*- sml -*-
 * nesting.sml
 *
 * COPYRIGHT (c) 2001 Bell Labs, Lucent Technologies
 *)

structure Nesting : sig

    (* val maxChunkSz : int ref *)

    val doit : int -> Document.doc -> Document.doc

  end = struct

    structure D = Document

  (* this parameter controls the maximum length of a list that we will apply
   * the dynamic programming to.
   *)
    (* val maxChunkSz = ref 32 *)

  (* subscript operations *)
#include "sub.ins"

    datatype tree
      = LF of {
	  tags : D.tagging,
	  ign : D.ignore,
	  index : int
	}
      | PAIR of {	    (* adjacent elements with disjoint tags *)
	  ign : D.ignore,
	  cost : int,		/* cost of this node's markup plus the kids */
	  l : tree,
	  r : tree
	}
      | MERGE of {	    (* merged tags of two or more adjacent elements *)
	  tags : D.tagging,	(* non-empty *)
	  ign : D.ignore,
	  cost : int,		/* cost of this node's markup plus the kids */
	  l : tree,
	  r : tree
	}

    val dummyTree = LF{
	    tags = OptUtil.nopTagging,
	    ign = {color = false, sz = false, font = 0w0},
	    index = ~1
	  }

  (* return the cost of a tagging *)
    fun taggingCost {font, color, sz} = let
	  val cost = ord(stringSub(Tables.taggingCostTbl, Word.toIntX font))
	  val cost = if (color >= 0) then cost + 7 else cost
	  val cost = if (sz >= 0) then cost + 7 else cost
	  in
	    cost
	  end

  (* get the ignored field of a tree *)
    fun ignOf (LF{ign, ...}) = ign
      | ignOf (PAIR{ign, ...}) = ign
      | ignOf (MERGE{ign, ...}) = ign

  (* get the cost of a tree *)
    fun cost (LF{tags, ...}) = taggingCost tags
      | cost (PAIR{cost, ...}) = cost
      | cost (MERGE{cost, tags, ...}) = cost

  (* pair two trees that cannot be merged *)
    fun pair (t1, t2) = PAIR{
	    cost = cost t1 + cost t2,
	    ign = FastIInfer.i_pair(ignOf t1, ignOf t2),
	    l = t1, r = t2
	  }

    fun i_merge (tags, t1, t2) =
	  FastIInfer.i_elem(tags, FastIInfer.i_pair(ignOf t1, ignOf t2))

  (* merge two trees *)
    fun merge (tags, tagCost, t1, t2) = MERGE{
	    tags = tags,
	    cost = cost t1 + cost t2 + tagCost,
	    ign = i_merge(tags, t1, t2),
	    l = t1, r = t2
	  }

    fun factor (tagging1, tagging2) = let
	  val {common, left, right} = Factor.factor (tagging1, tagging2)
	  in
	    if OptUtil.isNop common
	      then NONE
	      else SOME(common, left, right)
	  end

    exception AbortMerge

    fun i_safe (tags, ign) = let
	  val tags = FastIInfer.i_safe(tags, ign)
	  in
	    if (OptUtil.isNop tags) then raise AbortMerge else tags
	  end

    fun findLeftTag (PAIR{l, ...}) = findLeftTag l
      | findLeftTag (LF{tags,...}) = tags
      | findLeftTag (MERGE{tags,l,...}) =
	OptUtil.collapseTagging(findLeftTag l, tags)

    fun findRightTag (PAIR{r, ...}) = findRightTag r
      | findRightTag (LF{tags,...}) = tags
      | findRightTag (MERGE{tags,r,...}) =
	OptUtil.collapseTagging(findRightTag r, tags)

(*
    fun findLeftTag (PAIR{l, r, ...}) = i_safe(findLeftTag l, ignOf r)
      | findLeftTag (LF{tags, ...}) = tags
      | findLeftTag (MERGE{tags, ...}) = tags

    fun findRightTag (PAIR{l, r, ...}) = i_safe(findRightTag r, ignOf l)
      | findRightTag (LF{tags, ...}) = tags
      | findRightTag (MERGE{tags, ...}) = tags
*)

(*
    fun leafResidual (tags, common) = #left (Factor.factor(tags, common))
*)

    fun updateTree (common, t) = let
	fun update (PAIR{l, r, ...},delta) =
	    let val l = update(l, delta)
		val r = update(r, delta)
	    in
		PAIR{ign=FastIInfer.i_pair(ignOf l, ignOf r),
		     cost = cost l + cost r,
		     l = l, r = r}
	    end
	  | update (LF{tags, index, ign}, delta) =
	    LF{tags = #left(Factor.factor(tags,delta)),
	       index=index, ign=ign}
	  | update (MERGE{tags,l,r,...},delta) =
	    let val {left,right,...} = Factor.factor (tags, delta)
		val l = update (l, right)
		val r = update (r, right)
	    in
		merge (left, taggingCost left, l, r)
	    end
    in
	update (t, common)
    end

(*
    fun updateLeftTag (common, t, savings) = let
	  fun update (PAIR{ign, cost, l, r}) =
		PAIR{ign=ign, cost=cost-savings, l=update l, r=r}
	    | update (LF{tags, ign, index}) =
		LF{tags=leafResidual (tags, common), ign=ign, index=index}
	    | update (MERGE{tags, ign, cost, l, r}) =
		MERGE{tags=leafResidual (tags, common), ign=ign, cost=cost-savings, l=l, r=r}
	  in
	    update t
	  end

    fun updateRightTag (common, t, savings) = let
	  fun update (PAIR{ign, cost, l, r}) =
		PAIR{ign=ign, cost=cost-savings, l=l, r=update r}
	    | update (LF{tags, ign, index}) =
		LF{tags=leafResidual (tags, common), ign=ign, index=index}
	    | update (MERGE{tags, ign, cost, l, r}) =
		MERGE{tags=leafResidual (tags, common), ign=ign, cost=cost-savings, l=l, r=r}
	  in
	    update t
	  end
*)

  (* catenate two trees; note that we do not have to keep ignore info current
   * below merge nodes, since the PAIR merge stops at MERGE nodes.
   *)
    fun cat (
	  t1 as LF{tags=tg1, ign=i1, index=ix1},
	  t2 as LF{tags=tg2, ign=i2, index=ix2}
	) = (
	  case factor(tg1, tg2)
	   of NONE => pair(t1, t2)
	    | SOME(common, res1, res2) => let
		val savings = taggingCost common
		val t1 = LF{tags=res1, ign=i1, index=ix1}
		val t2 = LF{tags=res2, ign=i2, index=ix2}
		in
		  merge (common, savings, t1, t2)
		end
	  (* end case *))
      | cat (
	  t1 as LF{tags=tg1, ign=i1, index=ix1},
	  t2 as MERGE{tags=tg2, ign=i2, cost=c2, l=l2, r=r2}
	) = (
	  case factor(tg1, tg2)
	   of NONE => pair(t1, t2)
	    | SOME(common, res1, res2) => let
		val savings = taggingCost common
		val t1 = LF{tags=res1, ign=i1, index=ix1}
		val t2 = MERGE{tags=res2, ign=i2, cost=c2-savings, l=l2, r=r2}
		in
		  merge (common, savings, t1, t2)
		end
	  (* end case *))
      | cat (
	  t1 as MERGE{tags=tg1, ign=i1, cost=c1, l=l1, r=r1},
	  t2 as LF{tags=tg2, ign=i2, index=ix2}
	) = (
	  case factor(tg1, tg2)
	   of NONE => pair(t1, t2)
	    | SOME(common, res1, res2) => let
		val savings = taggingCost common
		val t1 = MERGE{tags=res1, ign=i1, cost=c1-savings, l=l1, r=r1}
		val t2 = LF{tags=res2, ign=i2, index=ix2}
		in
		  merge (common, savings, t1, t2)
		end
	  (* end case *))
      | cat (
	  t1 as MERGE{tags=tg1, ign=i1, cost=c1, l=l1, r=r1},
	  t2 as MERGE{tags=tg2, ign=i2, cost=c2, l=l2, r=r2}
	) = (
	  case factor(tg1, tg2)
	   of NONE => pair(t1, t2)
	    | SOME(common, res1, res2) => let
		val savings = taggingCost common
		val t1 = MERGE{tags=res1, ign=i1, cost=c1-savings, l=l1, r=r1}
		val t2 = MERGE{tags=res2, ign=i2, cost=c2-savings, l=l2, r=r2}
		in
		  merge (common, savings, t1, t2)
		end
	  (* end case *))
      | cat (t1, t2) =
	let val tg1 = findLeftTag t1
	    val tg2 = findRightTag t2
	    val ign = FastIInfer.i_pair (ignOf t1, ignOf t2)
	    val common0 = #common (Factor.factor (tg1, tg2))
	    val common = FastIInfer.i_safe (common0, ign)
	in
	    if OptUtil.isNop common then pair (t1, t2)
	    else let val t1 = updateTree (common, t1)
		     val t2 = updateTree (common, t2)
		 in
		     merge (common, taggingCost common, t1, t2)
		 end
	end
(*
      | cat (t1, t2) = (let
	  val tg1 = findLeftTag t1
	  val tg2 = findRightTag t2
	  in
	    case factor(tg1, tg2)
	     of NONE => pair(t1, t2)
	      | SOME(common, res1, res2) => let
		  val savings = taggingCost common
		  val t1 = updateLeftTag (common, t1, savings)
		  val t2 = updateRightTag (common, t2, savings)
		  in
		    merge (common, savings, t1, t2)
		  end
	    (* end case *)
	  end handle AbortMerge => pair(t1, t2))
*)

    fun doDoc sz (D.LIST{body, ign}) = let
	  val chunks = prepareList(sz, body, [])
	  val result = applyOpt chunks
	  in
	    D.LIST{body= result, ign= FastIInfer.i_list (map D.ignOf result)}
	  end
      | doDoc sz (D.ELEM{tags, content, ign}) =
	  D.ELEM{tags=tags, content=doDoc sz content, ign=ign}
      | doDoc _ elem = elem

  (* prepareList flattens nested LIST elements and recursively applies
   * the optimization to the list elements.
   *)
    and prepareList (sz, [], l) = chunkList (sz, l, [])
      | prepareList (sz, D.LIST{body, ...}::r, l) = prepareList (sz ,body@r, l)
      | prepareList (sz, elem::r, l) = prepareList(sz, r, doDoc sz elem :: l)

  (* chunkList breaks the list into sublists that can be optimized; the
   * result is a list of lists in left to right order.
   *)
    and chunkList (sz, [] : D.doc list, ll : D.doc list list) = ll
      | chunkList (sz, e::r, ll) = let
	  fun gatherChunk (_, [], chunk) = chunk::ll
	    | gatherChunk (1, l, chunk) = chunkList (sz, l, chunk::ll)
	    | gatherChunk (i, e::r, chunk) = gatherChunk (i-1, r, e::chunk)
	  in
	    gatherChunk (sz, r, [e])
	  end

  (* the optimization pass works on one chunk at a time; it does dynamic
   * programming to determine the best tagging scheme for each chunk.
   *)
    and applyOpt [] = []
      | applyOpt ([e]::r) = e :: applyOpt r
      | applyOpt (chunk::r) = optChunk (chunk, applyOpt r)

  (* optimize a single chunk and add it to the result list.  It should
   * be the case that chunk has at least two elements.
   *)
    and optChunk (chunk, result) = let
	  val elems = Vector.fromList chunk
	  val nElems = Vector.length elems
	(* allocate a triangular table *)
	  fun mkLeaf j = let
		val (tags, ign) = (case vectorSub(elems, j)
		       of D.ELEM{tags, ign, ...} => (tags, ign)
			| (D.SP{ign} | D.TEXT{ign, ...}) =>
			    (OptUtil.nopTagging, ign)
		      (* end case *))
		in
		  LF{tags = tags, ign = ign, index = j}
		end
	  val tbl = Vector.tabulate(nElems,
		  fn 0 => Array.tabulate(nElems, mkLeaf)
		   | i => Array.array(nElems-i, dummyTree))
	  fun get {start, len} = arraySub(vectorSub(tbl, len-1), start)
	(* try joining [i..i+k-1] with [i+k..i+j-1]; where 1 <= k < j *)
	  fun try (i, k, j) = let
		val t1 = get{start=i, len=k}
		val t2 = get{start=i+k, len=j-k}
		in
		  cat(t1, t2)
		end
	(* compute low cost span for [i..i+j-1] *)
	  fun findBest (i, j) = let
		val firstTry = try(i, 1, j)
		fun lp (best, bestCost, k) = if (k < j)
		      then let
			val t = try(i, k, j)
			val c = cost t
			in
			  if (c < bestCost)
			    then lp (t, c, k+1)
			    else lp (best, bestCost, k+1)
			end
		      else best
		val best = lp (firstTry, cost firstTry, 2)
		in
		  arrayUpdate(vectorSub(tbl, j-1), i, best)
		end
	  fun dp (lvl, row) = let
		val n = Array.length row
		fun lp i = if (i < n)
		      then (findBest(i, lvl+1); lp(i+1))
		      else ()
		in
		  lp 0
		end
	  in
	    Vector.appi dp (tbl, 1, NONE);
	    saveResult (arraySub(vectorSub(tbl, nElems-1), 0), elems, result)
	  end

  (* convert the optimal tree back into a a document *)
    and saveResult (tr, elems, result) = let
	  fun cvt (LF{tags, ign, index}, es) = (case vectorSub(elems, index)
		 of D.ELEM{content, ...} =>
		      D.ELEM{tags=tags, ign=ign, content=content} :: es
		  | e => e::es
		(* end case *))
	    | cvt (PAIR{l, r, ...}, es) = cvt(l, cvt(r, es))
	    | cvt (MERGE{tags, ign, l, r, ...}, es) =
		D.ELEM{
		    tags = tags, ign = ign,
		    content = D.LIST{
		        body = cvt(l, cvt(r, [])),
		        ign = FastIInfer.i_pair(ignOf l, ignOf r)
		      }
		  } :: es
	  in
	    cvt(tr, result)
	  end

    val doit = doDoc

  end
