(* -*- sml -*-
 * output.cpp
 *
 * COPYRIGHT (c) 2001 Bell Labs, Lucent Technologies
 *
 * Output routine for SML/NG documents.  This prints the representation produced
 * by the input routine.  The optimization takes place on the document structure,
 * without touching the underlying text (the string list).
 *)

structure Output : sig

    val output : Document.doc -> unit

  end = struct

    structure D = Document
    structure SIO = TextIO.StreamIO
    structure SS = Substring

#include "bits.ins"

    fun fontStrings font = let
	  fun addTag ((b, tag), tags) = if testBit b font
		then tag::tags
		else tags
	  val init = (case projBits u_bits font
		 of 0w0 => []
		  | 0w1 => ["U"]
		  | 0w2 => ["U", "U"]
		  | 0w3 => ["U", "U", "U"]
		(* end case *))
	  in
	    List.foldr addTag init [
		(pl_bit, "PL"),
		(b_bit, "B"),
		(i_bit, "I"),
		(tt_bit, "TT"),
		(s_bit, "S"),
		(em_bit, "EM")
	      ]
	  end

    fun tagsToStrings {font, sz, color} = let
	  val tags = fontStrings font
	  val tags = if (sz >= 0)
		then Int.toString sz :: tags
		else tags
	  val tags = if (color >= 0)
		then str(String.sub("rgbcmykw", color)) :: tags
		else tags
	  in
	    tags
	  end

    fun output doc = let
	  val outStrm = TextIO.getOutstream TextIO.stdOut
	  fun outText body = SIO.output(outStrm, body)
	(* we print spaces as newlines for the top-level content, since it makes
	 * the output a bit easier to read.
	 *)
	  fun topOut [] = ()
	    | topOut (D.SP{...}::r) = (SIO.output1(outStrm, #"\n"); topOut r)
	    | topOut (D.TEXT{body, ...}::r) = (outText body; topOut r)
	    | topOut (D.LIST{body, ...}::r) = (topOut body; topOut r)
	    | topOut (D.ELEM arg::r) = (outElem arg; topOut r)
	  and outElem {tags, content, ...} = let
		fun emit [] = out content
		  | emit (tag::r) = (
		      SIO.output(outStrm, String.concat["<", tag, ">"]);
		      emit r;
		      SIO.output(outStrm, String.concat["</", tag, ">"]))
		in
		  emit (tagsToStrings tags)
		end
	  and out (D.SP{...}) = SIO.output1(outStrm, #" ")
	    | out (D.TEXT{body, ...}) = outText body
	    | out (D.LIST{body, ...}) = List.app out body
	    | out (D.ELEM arg) = outElem arg
	  in
	    ignore (topOut [doc])
	  end

  end
