(* -*- sml -*-
 * td-contract.cpp
 *
 * (C) 2001 Lucent Technologies, Bell Labs
 *)
structure TDContract : sig
    (* contract:
     *    - eliminate tags that do nothing in the given context
     *    - eliminate tags that are ignored by their content
     *    - recalculate ign information
     *    - flatten LISTs
     *    - collapse directly nested ELEMs
     *)
    val contract : Document.doc -> Document.doc
end = struct

    structure D = Document
    structure II = FastIInfer
    structure OU = OptUtil

#include "bits.ins"

    infix || &&
    val op || = Word.orb
    val op && = Word.andb

    val bittsem_bits = b_bit || i_bit || tt_bit || s_bit || em_bit

    val f_ign_everything = #font II.i_everything

    fun reduceTagsInContext (tags, ctxt) = let

        val { color = t_color, sz = t_sz, font = t_font } = tags
        val { color = c_color, sz = c_sz, font = c_font } = ctxt

        val (t_color, c_color) =
            if c_color = t_color then (~1, c_color)
            else (t_color, t_color)

        val (t_sz, c_sz) =
            if c_sz = t_sz then (~1, c_sz)
            else (t_sz, t_sz)

        val (pl, c_font) =
            if testBit pl_bit t_font then (c_font <> 0w0, 0w0)
            else (false, c_font)

        fun doStyle (bit, c_font) =
            if testBit bit t_font then
                if testBit bit c_font then (false, c_font)
                else (true, setBit bit c_font)
            else (false, c_font)

        val (b, c_font) = doStyle (b_bit, c_font)
        val (i, c_font) = doStyle (i_bit, c_font)
        val (tt, c_font) = doStyle (tt_bit, c_font)
 
       val (s, c_font) =
            if testBit s_bit t_font then
                if testBit s_bit c_font then (false, c_font)
                else if testBit em_bit c_font then
                    (true, setBit s_bit (clearBit em_bit c_font))
                else (true, setBit s_bit c_font)
            else (false, c_font)

        val (em, c_font) =
            if testBit em_bit t_font then
                if testBit s_bit c_font then (false, c_font)
                else (true, flipBit em_bit c_font)
            else (false, c_font)

        val (u, c_font) = let
            val incr = projBits u_bits t_font
            val cur = projBits u_bits c_font
            val new = incr + cur
            val new = if new > 0w3 then 0w3 else new
            val need = new - cur
        in
            (need, setU (c_font, new))
        end

        fun bb _ false = 0w0
          | bb b true = b
        val t_font =
            bb pl_bit pl || bb b_bit b || bb i_bit i || bb tt_bit tt ||
            bb s_bit s || bb em_bit em || u
    in
        ({ color = t_color, sz = t_sz, font = t_font },
         { color = c_color, sz = c_sz, font = c_font })
    end

    fun elimIgnoredTags_fast (t : D.tagging, ig : D.ignore) = let
	val rt = II.i_strip (t, ig)
    in
	(rt, II.i_elem (rt, ig))
    end

    fun elimIgnoredTags_slow (tags: D.tagging, ign: D.ignore) = let
        val i_color = #color ign
        val i_sz = #sz ign
        val t_color = if i_color then ~1 else #color tags
        val t_sz = if i_sz then ~1 else #sz tags
        val t_font = #font tags
        val i_font = #font ign
        val (u, i_font) = let
            val i_cur = projBits u_bits i_font
            val t_try = projBits u_bits t_font
            val i_new = Word.max (i_cur - t_try, 0w0)
            val t_eff = i_cur - i_new
        in
            (t_eff, setU (i_font, i_new))
        end
        val em = testBit em_bit t_font andalso not (testBit em_bit i_font)
        val (s, i_font) =
            if testBit s_bit t_font then
                if testBit s_bit i_font then (false, i_font)
                else (true, setBit s_bit (setBit em_bit i_font))
            else (false, i_font)
        fun monotone (bit, i_font) =
            if testBit bit t_font then
                if testBit bit i_font then (false, i_font)
                else (true, setBit bit i_font)
            else (false, i_font)
        val (tt, i_font) = monotone (tt_bit, i_font)
        val (i, i_font) = monotone (i_bit, i_font)
        val (b, i_font) = monotone (b_bit, i_font)
        val (pl, i_font) =
            if testBit pl_bit t_font then
                if testBit pl_bit i_font then (false, i_font)
                else (true, f_ign_everything)
            else (false, i_font)
        fun bb _ false = 0w0
          | bb b true = b
        infix ||
        val op || = Word.orb
        val t_font =
            bb pl_bit pl || bb b_bit b || bb i_bit i || bb tt_bit tt ||
            bb s_bit s || bb em_bit em || u
    in ({ color = t_color, sz = t_sz, font = t_font },
        { color = i_color orelse t_color <> ~1,
	  sz = i_sz orelse t_sz <> ~1,
	  font = i_font })
    end

    fun elimIgnoredTags args = let
	val r_fast = elimIgnoredTags_fast args
#ifndef NDEBUG
	val r_slow = elimIgnoredTags_slow args
	val _ = if #1 r_fast = #1 r_slow then ()
		else (TextIO.output (TextIO.stdErr,
				    "!!! elimIgnoredTags tags discrepancy!\n");
		      OS.Process.exit OS.Process.failure)
	val _ = if #2 r_fast = #2 r_slow then ()
		else (TextIO.output (TextIO.stdErr,
				    "!!! elimIgnoredTags ign discrepancy!\n");
		      OS.Process.exit OS.Process.failure)
#endif
    in
	r_fast
    end

    val elimIgnoredTags = elimIgnoredTags_fast

    fun contract document = let
        fun opt _ (d as D.SP _) = d
          | opt _ (d as D.TEXT _) = d
          | opt ctxt (d as D.LIST { body, ign }) = let
		fun flatmapopt ([], l, i) =
		    D.LIST { body = rev l, ign = i }
		  | flatmapopt (h :: t, l, i) =
		    (case opt ctxt h of
			 D.LIST { body, ign } =>
			 flatmapopt (t, List.revAppend (body, l),
				     II.i_pair (i, ign))
		       | other =>
			 flatmapopt (t, other :: l,
				     II.i_pair (D.ignOf other, i)))
                val body2 = map (opt ctxt) body
            in
		flatmapopt (body, [], II.i_everything)
            end
          | opt ctxt (D.ELEM { tags, content, ign = _ }) = let
		fun collapseElems (t1, D.ELEM { tags = t2, content, ... }) =
		    collapseElems (OU.collapseTagging (t1, t2), content)
		  | collapseElems x = x
		val (tags, content) = collapseElems (tags, content)
                val (tags2, ctxt2) = reduceTagsInContext (tags, ctxt)
                val content2 = opt ctxt2 content
                val ign2 = D.ignOf content2
                val (tags3, ign3) = elimIgnoredTags (tags2, ign2)
            in
                if OU.isNop tags3 then content2
                else D.ELEM { tags = tags3, content = content2, ign = ign3 }
            end
        val start_context =
            { color = ~1, sz = ~1, font = 0w0 }
    in
        opt start_context document
    end
end
